/***********************************************************
* Rocks'n'Diamonds -- McDuffin Strikes Back!               *
*----------------------------------------------------------*
*               Artsoft Entertainment                      *
*               Holger Schemel                             *
*               Detmolder Strasse 189                      *
*               33604 Bielefeld                            *
*               Germany                                    *
*               e-mail: info@artsoft.org                   *
*----------------------------------------------------------*
* snoko.c                                                  *
*----------------------------------------------------------*
* (c) by the owner of the alias/nickname "HerzAusGold"     *
*               e-mail: herzausgold@online.de              *
***********************************************************/


#include "libgame/libgame.h"

#include "events.h"
#include "init.h"
#include "tools.h"
#include "libgame/util.h"
#include "libgame/mouse.h"
#include "snoko.h"

/*---*/

#define LINKS               left
#define RECHTS              right
#define RAUF                up
#define RUNTER              down

/*---*/

#define FELD(x,y)	Feld[x][y]

#define PLAYERX         pos_x
#define PLAYERY         pos_y

/*---*/

typedef short	     feldEle_t;

typedef struct {
	feldEle_t fd;
	int       x;
	int       y;
} feldEleDraw_t;

/*---*/

static short        findMap[MAX_LEV_FIELDX][MAX_LEV_FIELDY];

static short        pos_x = 0;
static short        pos_y = 0;
static short        left  = 0;
static short        right = 0;
static short        up    = 0;
static short        down  = 0;

static int          minx  = 0;
static int          miny  = 0;
static int          maxx  = MAX_LEV_FIELDX;
static int          maxy  = MAX_LEV_FIELDY;

/*---*/

static boolean  snap_state  = TRUE;
static int      old_button  = 0;
static int      spx         = -32;  /* outside screen */
static int      spy         = -32;  /* outside screen */


/**********************************************************/


#define BADMOVE   (MAX_LEV_FIELDX*MAX_LEV_FIELDY)


static int playerX = 0;
static int playerY = 0;


static void ClearMap()
{
  int x,y;
  for (  x=0; x<MAX_LEV_FIELDX; x++ )	{
    for (  y=0; y<MAX_LEV_FIELDY; y++ )	{
      findMap[x][y] = BADMOVE;
    }
  }
}

static void doStep(List_t *pList, short i_step, short i_push, short i_snap)
{
  step_t step;
  step.step = i_step;
  step.push = i_push;
  step.snap = i_snap;
  listTail(pList);
  listEnter(pList, &step, sizeof(step_t), FALSE);
}

static void FindTarget(int x, int y, int pathLen)
{	
  static int element = 0;
  if (x<minx || y<miny || x>=maxx || y>=maxy)  // 0, 0, MAX_LEV_FIELDX, MAX_LEV_FIELDY
    return;
  element = FELD(x,y);
  if ((element != EL_EMPTY)      && 
      (!IS_WALKABLE(element))    &&
      (!IS_DIGGABLE(element))    &&
      (!IS_COLLECTIBLE(element)) &&
      (!IS_SWITCHABLE(element))  &&
      (!IS_COLLECTIBLE_ONLY(element)) && 
      (!(x == playerX && y == playerY))   
     ) {
    return;
  }
  if (findMap[x][y] <= pathLen)
    return;
	
  findMap[x][y] = pathLen++;
		
  if((x == playerX) && (y == playerY))
    return;
	
  FindTarget(x - 1, y, pathLen);
  FindTarget(x + 1, y, pathLen);
  FindTarget(x, y - 1, pathLen);
  FindTarget(x, y + 1, pathLen);
}

static void FindTargetNoCheck(int x, int y, int pathLen)
{	
  static int element = 0;
  if (x<minx || y<miny || x>=maxx || y>=maxy)  // 0, 0, MAX_LEV_FIELDX, MAX_LEV_FIELDY
    return;
  element = FELD(x,y);
/*
  if ((element != EL_EMPTY)      && 
      (!IS_WALKABLE(element))    &&
      (!IS_DIGGABLE(element))    &&
      (!IS_COLLECTIBLE(element)) &&
      (!IS_COLLECTIBLE_ONLY(element)) && 
      (!(x == playerX && y == playerY))   
     ) {
    return;
  }
*/
  if (findMap[x][y] <= pathLen)
    return;
	
  findMap[x][y] = pathLen++;
		
  if((x == playerX) && (y == playerY))
    return;
	
  FindTargetNoCheck(x - 1, y, pathLen);
  FindTargetNoCheck(x + 1, y, pathLen);
  FindTargetNoCheck(x, y - 1, pathLen);
  FindTargetNoCheck(x, y + 1, pathLen);
}


static bool_t findPath (int ax, int ay)
{
  ClearMap();	
  FindTarget(ax, ay, 0);
  if (findMap[playerX][playerY]!=BADMOVE)
  {
    // save steps
    int cx = playerX;
    int cy = playerY;
    while (findMap[cx][cy])
    {
      if (findMap[cx - 1][cy] == (findMap[cx][cy] - 1)) {
        doStep(&goList, LINKS,  FALSE, FALSE);
        cx--;
      } else if (findMap[cx + 1][cy] == (findMap[cx][cy] - 1)) {
	doStep(&goList, RECHTS, FALSE, FALSE);
	cx++;
      } else if (findMap[cx][cy - 1] == (findMap[cx][cy] - 1)) {
        doStep(&goList, RAUF,   FALSE, FALSE); 
	cy--;
      } else if (findMap[cx][cy + 1] == (findMap[cx][cy] - 1)) {
        doStep(&goList, RUNTER, FALSE, FALSE);
	cy++;
      } else {
        break;  // failure
      } 
    }
  }
  return (!listEmpty(&goList));
}


static bool_t findPathNoCheck (int ax, int ay)
{
  ClearMap();	
  FindTargetNoCheck(ax, ay, 0);
  if (findMap[playerX][playerY]!=BADMOVE)
  {
    // save steps
    int cx = playerX;
    int cy = playerY;
    while (findMap[cx][cy])
    {
      if (findMap[cx - 1][cy] == (findMap[cx][cy] - 1)) {
        doStep(&goList, LINKS,  FALSE, FALSE);
        cx--;
      } else if (findMap[cx + 1][cy] == (findMap[cx][cy] - 1)) {
	doStep(&goList, RECHTS, FALSE, FALSE);
	cx++;
      } else if (findMap[cx][cy - 1] == (findMap[cx][cy] - 1)) {
        doStep(&goList, RAUF,   FALSE, FALSE); 
	cy--;
      } else if (findMap[cx][cy + 1] == (findMap[cx][cy] - 1)) {
        doStep(&goList, RUNTER, FALSE, FALSE);
	cy++;
      } else {
        break;  // failure
      } 
    }
  }
  return (!listEmpty(&goList));
}


void snoko (int mode, int ax, int ay,int posx, int posy, 
            short le, short ri, short ups, short dwn)
{
   bool_t find; 
   step_t myStep;

   pos_x = posx;
   pos_y = posy;
   left  = le;
   right = ri;
   up    = ups;
   down  = dwn;   

   if (ax != PLAYERX || ay != PLAYERY) 
   {	  	
     // player
     playerX = PLAYERX;
     playerY = PLAYERY;	  	    

     if (mode == SNOKOMODE_DIRECT) 
     {
       int    dx,dy;
       int    i;
       dx = ax - playerX; 
       dy = ay - playerY; 
       if (dx || dy) {
         if (abs(dx) > abs(dy)) {
           if (dx < 0) {
             myStep.push = 0;
             myStep.snap = 0;
             myStep.step = MV_LEFT;
           } else {
             myStep.push = 0;
             myStep.snap = 0;
             myStep.step = MV_RIGHT;
           }
           if (options.mouseOption.followMouse) {
             if (listLength(&goList) <= 0) {
               listTail(&goList);
               listEnter(&goList,&myStep,sizeof(step_t),FALSE);
             }
           } else {
             listTail(&goList);
             for (i=0; i<abs(dx); i++) {
               listEnter(&goList,&myStep,sizeof(step_t),FALSE);
             }
           }
         } else {
           if (dy < 0) {
             myStep.push = 0;
             myStep.snap = 0;
             myStep.step = MV_UP;
           } else {
             myStep.push = 0;
             myStep.snap = 0;
             myStep.step = MV_DOWN;
           }
           if (options.mouseOption.followMouse) {
             if (listLength(&goList) <= 0) {
               listTail(&goList);
               listEnter(&goList,&myStep,sizeof(step_t),FALSE);
             }
           } else {
             listTail(&goList);
             for (i=0; i<abs(dy); i++) {
               listEnter(&goList,&myStep,sizeof(step_t),FALSE);
             }
           }
         }
       }

     } 
     else if (mode == SNOKOMODE_SMALL)
     {
       // min 
       minx = playerX - 4 - 1;
       if (minx < 0) minx = 0;
       miny = playerY - 4 - 1;
       if (miny < 0) miny = 0;

       // max
       maxx = playerX + 4 + 1;
       if (maxx >= MAX_LEV_FIELDX) maxx = MAX_LEV_FIELDX;
       maxy = playerY + 4 + 1;
       if (maxy >= MAX_LEV_FIELDY) maxy = MAX_LEV_FIELDY;

       // steps 
       find = findPathNoCheck(ax, ay);
     } else {
       // min 
       minx = playerX - MIDPOSX - 1;
       if (minx < 0) minx = 0;
       miny = playerY - MIDPOSY - 1;
       if (miny < 0) miny = 0;

       // max
       maxx = playerX + MIDPOSX + 1;
       if (maxx >= MAX_LEV_FIELDX) maxx = MAX_LEV_FIELDX;
       maxy = playerY + MIDPOSY + 1;
       if (maxy >= MAX_LEV_FIELDY) maxy = MAX_LEV_FIELDY;

       // steps 	
       find = findPath(ax, ay);
     }

   } else {
     if (!options.mouseOption.learn) {
       listFree(&goList);
       listInit(&goList);
     } else {
       myStep.push = 0;
       myStep.snap = 0;
       myStep.step = 0;
       listTail(&goList);
       listEnter(&goList,&myStep,sizeof(step_t),FALSE);
     }
   } 
}  // snoko	  


/*-----------------------------------------------------------------------
 MouseHandleMove - called after HandleMotionEvent
-----------------------------------------------------------------------*/
void MouseHandleMove(int mx, int my)
{
  int ax,ay,dx,dy;
  step_t myStep;

  if (options.mouseOption.mouseHandleMove == MHM_DIRDIFF) 
  {   // MHM_DIRDIFF
    if (IN_GFX_SCREEN(mx,my)) {
      ax = LEVELX((mx-SX)/TILEX);
      ay = LEVELY((my-SY)/TILEY);

      dx = ax - stored_player[0].jx;
      dy = ay - stored_player[0].jy;
      if (dx || dy) {
        if (abs(dx) > abs(dy)) {
          if (dx < 0) {
            myStep.push = 0;
            myStep.snap = 0;
            myStep.step = MV_LEFT;
          } else {
            myStep.push = 0;
            myStep.snap = 0;
            myStep.step = MV_RIGHT;
          }
        } else {
          if (dy < 0) {
            myStep.push = 0;
            myStep.snap = 0;
            myStep.step = MV_UP;
          } else {
            myStep.push = 0;
            myStep.snap = 0;
            myStep.step = MV_DOWN;
          }
        }
        if (listLength(&goList) <= 0) {
          listTail(&goList);
          listEnter(&goList,&myStep,sizeof(step_t),FALSE);
        }
      }
    }    	
  } 

  if (options.mouseOption.mouseHandleMove == MHM_SNOKO) 
  {   // MHM_SNOKO
    if (IN_GFX_SCREEN(mx,my)) {
      ax = LEVELX((mx-SX)/TILEX);
      ay = LEVELY((my-SY)/TILEY);

      listFree(&goList);
      listInit(&goList);
      snoko(options.mouseOption.snokoMode, ax,ay,
            stored_player[0].jx,stored_player[0].jy,
	    MV_LEFT, MV_RIGHT, MV_UP, MV_DOWN);
    }
  }   // MHM_SNOKO

} // MouseHandleMove


/*-----------------------------------------------------------------------
 MouseHandleButton - called after HandleKeyEvent
-----------------------------------------------------------------------*/
void MouseHandleButton(int mx, int my, int button)
{
  int ax,ay,dx,dy;
  step_t myStep;

  if (button == MB_LEFTBUTTON && !IN_GFX_SCREEN(mx,my)) {
    snap_state = TRUE;
    old_button = 0;
    spx        = -32;     /* outside screen */
    spy        = -32;     /* outside screen */
    listFree(&goList);    /* Click Outside  */
    listInit(&goList);    /* Click Outside  */
    listFree(&learnList); /* Click Outside  */
    listInit(&learnList); /* Click Outside  */
  }

  if (options.mouseOption.mouseHandleButton == MHB_DIRSNAP) 
  {   // MHB_DIRSNAP
    if (options.mouseOption.followMouse) {
      if (button == MB_MIDDLEBUTTON && IN_GFX_SCREEN(mx,my)) {
        snap_state = !snap_state;
      }
      if (button == MB_LEFTBUTTON && IN_GFX_SCREEN(mx,my)) {
        if (snap_state) {
          myStep.step = MV_LEFT  | KEY_BUTTON_1;
          myStep.push = 0;
          myStep.snap = 1;
          listTail(&goList);
          listEnter(&goList, &myStep, sizeof(step_t), FALSE);
        } else {
          myStep.step = MV_UP    | KEY_BUTTON_1;
          myStep.push = 0;
          myStep.snap = 1;
          listTail(&goList);
          listEnter(&goList, &myStep, sizeof(step_t), FALSE);
        }
      }
      if (button == MB_RIGHTBUTTON && IN_GFX_SCREEN(mx,my)) {
        if (snap_state) {
          myStep.step = MV_RIGHT | KEY_BUTTON_1;
          myStep.push = 0;
          myStep.snap = 1;
          listTail(&goList);
          listEnter(&goList, &myStep, sizeof(step_t), FALSE);
        } else {
          myStep.step = MV_DOWN  | KEY_BUTTON_1;
          myStep.push = 0;
          myStep.snap = 1;
          listTail(&goList);
          listEnter(&goList, &myStep, sizeof(step_t), FALSE);
        }
      }
    }
  }

  if (options.mouseOption.mouseHandleButton == MHB_SOKO_SNOKO) 
  {   // MHB_SOKO_SNOKO
    boolean rmove;

    if (button == MB_MIDDLEBUTTON && IN_GFX_SCREEN(mx,my)) {
      if (!listEmpty(&learnList)) {
        listFree(&goList);
        listInit(&goList);
        listCopy(&learnList, &goList, sizeof(step_t));
      } 
    }

    if (options.mouseOption.learn) {
      /* list is empty - or outside screen */
      if (listEmpty(&goList) || (spx == -32 || spy == -32)) {
        spx = stored_player[0].jx;
        spy = stored_player[0].jy;
      }
    } else {
      spx = stored_player[0].jx;
      spy = stored_player[0].jy;
    }

    // checking left button
    if (button == MB_LEFTBUTTON && !motion_status && IN_GFX_SCREEN(mx,my)) {
      old_button = button;
    }
    if (button == MB_RELEASED && old_button == MB_LEFTBUTTON) {
      ax = LEVELX((mx-SX)/TILEX);
      ay = LEVELY((my-SY)/TILEY);
      old_button = 0;
      soko(ax,ay,
           spx,spy, 
           MV_LEFT, MV_RIGHT, MV_UP, MV_DOWN);
      if (options.mouseOption.learn) {
        spx = ax;
        spy = ay;
      }
    }
    // checking right button
    if (button == MB_RIGHTBUTTON && !motion_status && IN_GFX_SCREEN(mx,my)) {
      old_button = button;
    }
    rmove = 0;
    if (options.mouseOption.followMouse) {
      rmove = (button == MB_RIGHTBUTTON && IN_GFX_SCREEN(mx,my));
    } else {
      rmove = (button == MB_RELEASED && old_button == MB_RIGHTBUTTON) ;
    }
    if (rmove) {
      ax = LEVELX((mx-SX)/TILEX);
      ay = LEVELY((my-SY)/TILEY);
      old_button = 0;

      if (options.mouseOption.followMouse) {
        listFree(&goList);
        listInit(&goList);
      }
      snoko(options.mouseOption.snokoMode, ax,ay,
            spx,spy, 
            MV_LEFT, MV_RIGHT, MV_UP, MV_DOWN);
      if (options.mouseOption.learn) {
        spx = ax;
        spy = ay;
      }
    }
  }  // MHB_SOKO_SNOKO

  if (options.mouseOption.mouseHandleButton == MHB_SNOKOSNAP) 
  {   // MHB_SNOKOSNAP
    boolean lmove;

    if (button == MB_MIDDLEBUTTON && IN_GFX_SCREEN(mx,my)) {
      if (!listEmpty(&learnList)) {
        listFree(&goList);
        listInit(&goList);
        listCopy(&learnList, &goList, sizeof(step_t));
      } 
    }

    if (options.mouseOption.learn) {
      /* list is empty - or outside screen */
      if (listEmpty(&goList) || (spx == -32 || spy == -32)) {
        spx = stored_player[0].jx;
        spy = stored_player[0].jy;
      }
    } else {
      spx = stored_player[0].jx;
      spy = stored_player[0].jy;
    }

    // checking left button
    if (button == MB_LEFTBUTTON && !motion_status && IN_GFX_SCREEN(mx,my)) {
      old_button = button;
    }
    lmove = 0;
    if (options.mouseOption.followMouse) {
      lmove = (button == MB_LEFTBUTTON && IN_GFX_SCREEN(mx,my));
    } else {
      lmove = (button == MB_RELEASED && old_button == MB_LEFTBUTTON) ;
    }
    if (lmove) {
      ax = LEVELX((mx-SX)/TILEX);
      ay = LEVELY((my-SY)/TILEY);
      old_button = 0;

      if (options.mouseOption.followMouse) {
        listFree(&goList);
        listInit(&goList);
      }
      snoko(options.mouseOption.snokoMode, ax,ay,
            spx,spy,
            MV_LEFT, MV_RIGHT, MV_UP, MV_DOWN);
      if (options.mouseOption.learn) {
        spx = ax;
        spy = ay;
      }
    }
    // checking right button
    if (button == MB_RIGHTBUTTON && !motion_status && IN_GFX_SCREEN(mx,my)) {
      old_button = button;
    }
    if (button == MB_RELEASED && old_button == MB_RIGHTBUTTON) {
      ax = LEVELX((mx-SX)/TILEX);
      ay = LEVELY((my-SY)/TILEY);
      old_button = 0;

      dx = ax - spx;
      dy = ay - spy;
      if (dx || dy) {
        if (abs(dx) > abs(dy)) {
          if (dx < 0) {
            myStep.push = 0;
            myStep.snap = 1;
            myStep.step = KEY_BUTTON_1 | MV_LEFT;
          } else {
            myStep.push = 0;
            myStep.snap = 1;
            myStep.step = KEY_BUTTON_1 | MV_RIGHT;
          }
        } else {
          if (dy < 0) {
            myStep.push = 0;
            myStep.snap = 1;
            myStep.step = KEY_BUTTON_1 | MV_UP;
          } else {
            myStep.push = 0;
            myStep.snap = 1;
            myStep.step = KEY_BUTTON_1 | MV_DOWN;
          }
        }
        listTail(&goList);
        listEnter(&goList,&myStep,sizeof(step_t),FALSE);
      } else {
        myStep.push = 0;
        myStep.snap = 0;
        myStep.step = KEY_BUTTON_2;
        listTail(&goList);
        listEnter(&goList,&myStep,sizeof(step_t),FALSE);
      }
    }
  }

} // MouseHandleButton

